import ImageDetection from './ImageDetection.vue'
import ScatterPlot from './ScatterPlot.vue'
import ParamsWidget from './ParamsWidget.vue'
import FilterOptionsWidget from './FilterOptionsWidget.vue'
import FilterOperatorWidget from './FilterOperatorWidget.vue'

export default {
  imageDetection: ImageDetection,
  scatterPlot: ScatterPlot,
  paramsWidget: ParamsWidget,
  filterOptionsWidget: FilterOptionsWidget,
  filterOperatorWidget: FilterOperatorWidget
}
